/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 

package de.fzj.unicore.uas.security;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.net.ssl.SSLContext;

import org.apache.log4j.Logger;

import eu.unicore.security.util.KeystoreChecker;
import eu.unicore.security.util.Log;
import eu.unicore.security.util.client.ETDClientSettings;
import eu.unicore.security.util.client.IClientProperties;

/**
 * security properties
 *  
 * @author schuller
 */
public class ClientProperties implements IClientProperties {
	

	/**
	 * for CLIENT use: space separated name(s) of any extra UAS outbound security handler class(es)
	 */
	public final static String UAS_OUTHANDLER_NAME="uas.security.out.handler.classname";
	
	/**
	 * for CLIENT use: space-separated names of any extra UAS incoming security handler classes
	 */
	public final static String UAS_INHANDLER_NAME="uas.security.in.handler.classname";
	
	/**
	 * property defining whether SSL is enabled
	 */
	public static final String WSRF_SSL = "unicore.wsrflite.ssl";

	/**
	 * property defining the SSL keystore
	 */
	public static final String WSRF_SSL_KEYSTORE = "unicore.wsrflite.ssl.keystore";

	/**
	 * property defining the SSL keystore password
	 */
	public static final String WSRF_SSL_KEYPASS = "unicore.wsrflite.ssl.keypass";

	/**
	 * property defining the SSL keystore type (e.g. JKS)
	 */
	public static final String WSRF_SSL_KEYTYPE = "unicore.wsrflite.ssl.keytype";

	/**
	 * property defining the alias of the key to use
	 */
	public static final String WSRF_SSL_KEYALIAS = "unicore.wsrflite.ssl.keyalias";

	/**
	 * property defining the SSL trust store
	 */
	public static final String WSRF_SSL_TRUSTSTORE = "unicore.wsrflite.ssl.truststore";

	/**
	 * property defining the SSL trust tore password
	 */
	public static final String WSRF_SSL_TRUSTPASS = "unicore.wsrflite.ssl.truststorepass";

	/**
	 * property defining the SSL trust store type (e.g. JKS)
	 */
	public static final String WSRF_SSL_TRUSTTYPE = "unicore.wsrflite.ssl.truststoretype";

	
	private static final Logger logger=Log.getLogger(Log.SECURITY,ClientProperties.class);
	
	private final Properties properties;
	
	private ETDClientSettings etdSettings = new ETDClientSettings();
	
	private X509Certificate[] certChain;
	
	private PrivateKey privateKey;
	
	private boolean signMessage=true;
	
	private Map<String,Object>extraSecurityTokens;
	
	private ClassLoader classLoader;
	
	public ClientProperties(){
		this(new Properties());
	}
	
	public ClientProperties(Properties properties){
		super();
		this.properties=properties;
	}
	
	public ClientProperties(InputStream is)throws IOException{
		this();
		properties.load(is);
	}

	public ClientProperties(String propertiesFileName)throws IOException{
		this(new FileInputStream(propertiesFileName));
	}
	
	public ClientProperties clone(){
		if(logger.isTraceEnabled())logger.trace("Copying security props: "+toString());
		try{
			ClientProperties clone=(ClientProperties)super.clone();
			clone.properties.putAll(this.properties);
			clone.etdSettings = etdSettings.clone();
			if(extraSecurityTokens!=null){
				clone.extraSecurityTokens=new HashMap<String,Object>();
				clone.extraSecurityTokens.putAll(extraSecurityTokens);
			}
			return clone;
		}catch(CloneNotSupportedException cne){
			//can't happen
			return null;
		}
	}
	
	/**
	 * return named property. 
	 */
	protected String doGetProperty(String key) {
		String p=(String)properties.get(key);
		return p;
	}
	
	public boolean doHttpAuthn() {
		return false;
	}
	
	public boolean doSignMessage() {
		return signMessage;
	}


	public boolean doSSLAuthn()
	{
		return isSslEnabled();
	}

	public synchronized X509Certificate[] getCertificateChain() {
		if(certChain==null){
			readKeystore();
		}
		return certChain;
	}
	
	public ClassLoader getClassLoader() {
		return classLoader;
	}
	
	public synchronized ETDClientSettings getETDSettings() {
		if(certChain==null){
			readKeystore();
		}
		return etdSettings;
	}

	public Map<String, Object> getExtraSecurityTokens() {

		return extraSecurityTokens;
	}

	public Properties getExtraSettings() {
		return properties;
	}
	

	public String getHttpPassword() {
		return null;
	}
	
	public String getHttpUser() {
		return null;
	}
	
	public String getInHandlerClassNames() {
		return doGetProperty(UAS_INHANDLER_NAME);
	}
	
	public String getKeystore() {
		return doGetProperty(WSRF_SSL_KEYSTORE);
	}
	
	public String getKeystoreAlias() {
		return doGetProperty(WSRF_SSL_KEYALIAS);
	}

	
	public String getKeystoreKeyPassword() {
		return doGetProperty(WSRF_SSL_KEYPASS);
	}

	public String getKeystorePassword() {
		return doGetProperty(WSRF_SSL_KEYPASS);
	}
	
	public String getKeystoreType() {
		return doGetProperty(WSRF_SSL_KEYTYPE);
	}
	
	public String getOutHandlerClassNames() {
		return doGetProperty(UAS_OUTHANDLER_NAME);
	}

	/**
	 * returns the private key entry identified by getKeystoreAlias() from the keystore 
	 */
	public synchronized PrivateKey getPrivateKey() {
		if(privateKey==null){
			readKeystore();
		}
		return privateKey;
	}
	
	public String getProperty(String key){
		return doGetProperty(key);
	}
	
	public X509Certificate getPublicKey() {
		return getCertificateChain()[0];
	}
	
	public SSLContext getSSLContext() {
		return null;
	}
	
	public String getTruststore() {
		String result = doGetProperty(WSRF_SSL_TRUSTSTORE);
		// fallback to keystore when truststore is not set
		return result == null ? getKeystore() : result;
	}
	
	public String getTruststorePassword() {
		String result = doGetProperty(WSRF_SSL_TRUSTPASS);
		// fallback to keystore pass when truststore pass is not set
		if(result == null){
			if(getKeystore() != null && getKeystore().equals(getTruststore())){
				return getKeystorePassword();
			}
			else return null;
		}
		else{
			 return result;
		}
	}

	public String getTruststoreType() {
		String result = doGetProperty(WSRF_SSL_TRUSTTYPE);
		// fallback to keystore type when truststore type is not set
		if(result == null){
			if(getKeystore() != null && getKeystore().equals(getTruststore())){
				return getKeystoreType();
			}
			else return null;
		}
		else{
			 return result;
		}
	}
	
	/**
	 * returns true if SSL mode is enabled.<br/>
	 * SSL can be enabled by setting a property 
	 * "unicore.wsrflite.ssl" to "true" in the wsrflite.xml file 
	 * or by defining a system property
	 * @return
	 */
	public boolean isSslEnabled(){
		return Boolean.parseBoolean(doGetProperty(WSRF_SSL));
	}
	
	/**
	 * read the keys from the keystore
	 * @throws RuntimeException in case of IO or keystore errors
	 */
	private synchronized void readKeystore(){
			String keystoreAlias=getKeystoreAlias();
	    	String keystoreName = getKeystore();
	    	String keystoreType = getKeystoreType();
	    	String keyPassword = getKeystoreKeyPassword();
	    	try{
	    		KeyStore keyStore =load(keystoreName,keyPassword,keystoreType);
	    		if(keystoreAlias==null){
	    			keystoreAlias = KeystoreChecker.findAlias(keyStore);
		    		if(keystoreAlias==null){
		    			throw new IllegalArgumentException("Keystore "+keystoreName+" does not contain any key entries!");
		    		}
	    			logger.debug("No alias supplied, loading  <"+keystoreAlias+">");
	    		}
	    		else {
	    			logger.debug("Loading  <"+keystoreAlias+">");
	    		}
	    		Certificate[] path=keyStore.getCertificateChain(keystoreAlias);
	    		if(path==null)throw new IllegalArgumentException("Alias <"+keystoreAlias+"> cannot be found in keystore. Please check your configuration.");
	    		privateKey=(PrivateKey)keyStore.getKey(keystoreAlias, keyPassword.toCharArray());
	    		if(privateKey==null)throw new IllegalArgumentException("Alias <"+keystoreAlias+"> does not denote a key entry. Please check your configuration.");
	    		
	    		certChain=new X509Certificate[path.length];
	    		for(int i=0;i<path.length;i++){
	    			certChain[i]=(X509Certificate)path[i];
	    		}
	    		etdSettings.setIssuerCertificateChain(certChain);
	    		
	    	}catch(Exception e){ 
	    		//always deadly, so we will throw a runtime exeption
	    		throw new RuntimeException("Could not load certificate(s) from keystore.",e);
	    	}
	}
	
	/**
	 * load keystore
	 * @param name - the keystore file name
	 * @param password - the keystore password
	 * @param type - the type. If this is <code>null</code>, it will be attempted to infer the 
	 *               type from the filename. If this is impossible, jks and p12 will be attempted
	 * @return
	 * @throws Exception
	 */
	private KeyStore load(String name, String password, String type)throws IOException, CertificateException, KeyStoreException, NoSuchAlgorithmException{
		logger.debug("Reading from keystore: " + name);
		File f = new File(name);
		if(!f.exists())throw new IOException("File not found : "+name);
		if(type==null){
			String fName=f.getName().toLowerCase();
			if(fName.endsWith(".jks")){
				type="jks";
			}
			else if(fName.endsWith(".p12") || fName.endsWith(".pkcs12")){
				type="pkcs12";
			}
		}
		KeyStore keyStore=null;
		//first try with given type or JKS
		String tryType=type!=null?type:"jks";
		FileInputStream fis=new FileInputStream(f);
		try{
			keyStore = KeyStore.getInstance(tryType);
			keyStore.load(fis, password.toCharArray());
		}catch(IOException ex){
			if(type!=null){
				throw ex;
			}
			//attempt p12
			try{fis.close();}catch(Exception ignored){}
			fis=new FileInputStream(f);
			keyStore = KeyStore.getInstance("PKCS12");
			keyStore.load(fis, password.toCharArray());
		}
		finally{
			try{fis.close();}catch(Exception ignored){}
		}
		logger.debug("Keystore: " + name + " successfully loaded");
		return keyStore;
	}
	
	public void setClassLoader(ClassLoader classLoader) {
		this.classLoader = classLoader;
	}
	
	public void setExtraSecurityTokens(Map<String, Object> extraSecurityTokens) {
		this.extraSecurityTokens = extraSecurityTokens;
	}
	
	public void setInHandlerClassNames(String spaceSeparatedClassnames) {
		setProperty(UAS_INHANDLER_NAME,spaceSeparatedClassnames);
	}
	
	public void setKeystore(String path) {
		setProperty(WSRF_SSL_KEYSTORE,path);
	}
	
	public void setKeystoreAlias(String alias) {
		setProperty(WSRF_SSL_KEYALIAS,alias);
	}
	
	public void setKeystorePassword(String passwd) {
		setProperty(WSRF_SSL_KEYPASS,passwd);
	}
	
	public void setKeystoreType(String type) {
		setProperty(WSRF_SSL_KEYTYPE, type);
	}
	
	public void setOutHandlerClassNames(String spaceSeparatedClassnames) {
		setProperty(UAS_OUTHANDLER_NAME,spaceSeparatedClassnames);
	}
	
	/**
	 * set a property
	 * @param key
	 * @param value
	 */
	public void setProperty(String key, String value){
		if(value==null)properties.remove(key);
		else properties.put(key,value);
	}
	
	public void setSignMessage(boolean signMessage) {
		this.signMessage = signMessage;
	}
	
	public void setSslEnabled(boolean useSSL){
		setProperty(WSRF_SSL, ""+useSSL);
	}
	
	public void setTruststore(String path) {
		setProperty(WSRF_SSL_TRUSTSTORE,path);
	}
	
	public void setTruststorePassword(String passwd) {
		setProperty(WSRF_SSL_TRUSTPASS,passwd);
	}

	public void setTruststoreType(String type) {
		setProperty(WSRF_SSL_TRUSTTYPE, type);
	}

	public String toString(){
		String lineSep=System.getProperty("line.separator");
		StringBuffer sb=new StringBuffer();
		sb.append(super.toString()).append(lineSep);
		sb.append("Keystore <").append(doGetProperty(WSRF_SSL_KEYSTORE)).append(lineSep);
		sb.append("Truststore <").append(doGetProperty(WSRF_SSL_TRUSTSTORE)).append(lineSep);
		try{
			sb.append("Identity: ").append(getCertificateChain()[0].getSubjectX500Principal().getName()).append(lineSep);
		}catch(Exception e){
			sb.append("User not set.").append(lineSep);
		}
		sb.append("Signing messages: ").append(doSignMessage()).append(lineSep);
		sb.append("Extend trust delegation: ").append(etdSettings.isExtendTrustDelegation()).append(lineSep);
		return sb.toString();
	}	
}
