/*********************************************************************************
 * Copyright (c) 2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 
package de.fzj.unicore.uas.security;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;
import org.codehaus.xfire.MessageContext;
import org.codehaus.xfire.client.Client;
import org.codehaus.xfire.handler.AbstractHandler;
import org.codehaus.xfire.handler.Phase;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import eu.unicore.security.util.Log;
import eu.unicore.security.util.client.IClientProperties;
import eu.unicore.security.xfireutil.client.Configurable;
import eu.unicore.security.xfireutil.client.DSigOutHandler;

/**
 * If needed then a token with a proxy cert in PEM format is inserted into the header.<br/>
 * The properties of the issued proxy can be configured in the security properties passed
 * to the doInit() method <br/>
 * <b>This handler is intended for client use</b><br/>
 * 
 * @author schuller
 */
public class ProxyCertOutHandler extends AbstractHandler implements Configurable {
	
	private static final Logger logger = Log.getLogger(Log.SECURITY,ProxyCertOutHandler.class);
	
	protected IClientProperties sec;
	
	private Element proxyHeaderAsJDOM=null;
	

	/**
	 * property for defining the lifetime in seconds (default: 2*3600, i.e. two hours)
	 */
	public static final String PROXY_LIFETIME="unicore.proxy.lifetime";
	
	/**
	 * property for defining the keysize (default: 1024)
	 */
	public static final String PROXY_KEYSIZE="unicore.proxy.keysize";
	
	private long expiryInstant;
	
	private int lifetime=7200;
	
	public ProxyCertOutHandler() {
		super();
		setPhase(Phase.POST_INVOKE);
		before(DSigOutHandler.class.getName());
	}


	public synchronized void configure(IClientProperties sec) {
		this.sec = sec;
		ProxyCertificate proxy=null;
		if(!isProxyDN(sec.getPublicKey().getSubjectX500Principal().getName())){
			//generate a new proxy cert
			try{
				String lifetimeP=sec.getExtraSettings().getProperty(PROXY_LIFETIME);
				if(lifetimeP!=null){
					lifetime=Integer.parseInt(lifetimeP);
				}
				String keysizeP=sec.getExtraSettings().getProperty(PROXY_KEYSIZE);
				int keysize=1024;
				if(keysizeP!=null){
					keysize=Integer.parseInt(keysizeP);	
				}
				proxy=ProxyCertificate.getInstance(sec, false, lifetime*1000, keysize);
			}catch(Exception ex){
				logger.error("Can't generate proxy certificate",ex);
				return;
			}
		}
		if(proxy==null){
			logger.debug("Not generating proxy.");
			return;
		}
	
		try{
			StringBuilder sb=new StringBuilder();
			sb.append("<proxy:Proxy xmlns:proxy=\"http://www.unicore.eu/unicore6\">");
			sb.append(proxy.getPEMEncoded());
			sb.append("</proxy:Proxy>");
			proxyHeaderAsJDOM=new SAXBuilder().build(new ByteArrayInputStream(sb.toString().getBytes())).detachRootElement();
			expiryInstant=System.currentTimeMillis()+1000*lifetime;
		}
		catch(Exception ce){
			logger.fatal("Can't create Proxy header: " , ce);
		}
	
		if(logger.isDebugEnabled()){
			logger.debug("(Re-)initialised Proxy Outhandler");
			try{
				ByteArrayOutputStream bos=new ByteArrayOutputStream();
				new XMLOutputter(Format.getPrettyFormat()).output(proxyHeaderAsJDOM, bos);
				logger.debug(bos.toString());
			}catch(Exception e){
				logger.warn("",e);
			}
		}
	}

	public Element getProxyHeader() {
		return proxyHeaderAsJDOM;
	}
	
	public synchronized void invoke(MessageContext context) throws Exception {
		if (proxyHeaderAsJDOM == null){
			return;
		}
		
		//do nothing if not client call
		Boolean clientMode = (Boolean) context.getProperty(Client.CLIENT_MODE);
		if (clientMode == null || !clientMode.booleanValue())
			return;

		if(System.currentTimeMillis()>expiryInstant){
			configure(sec);
		}
		Element h = context.getOutMessage().getOrCreateHeader();
		h.addContent((Element)proxyHeaderAsJDOM.clone());
	}

	//pattern for checking if a DN represents a proxy cert
	private final static String proxyRE= "(CN=([\\d]+))|(CN=proxy)";
	private final static Pattern pattern = Pattern.compile(proxyRE,Pattern.CASE_INSENSITIVE); 
	
	private boolean isProxyDN(String dn) {
		return pattern.matcher(dn).find();
	}

}


