/*
 * Copyright (c) 2000-2004, Intel Corporation All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Intel Corporation nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package de.fzj.unicore.uas.security;

import java.io.ByteArrayInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.StringWriter;
import java.io.Writer;
import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Random;
import java.util.Set;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1OctetString;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.ASN1Set;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DERPrintableString;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.DERSet;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.jce.provider.X509CertificateObject;
import org.bouncycastle.openssl.PEMWriter;
import org.bouncycastle.x509.X509V3CertificateGenerator;

import eu.unicore.security.util.Log;
import eu.unicore.security.util.client.IClientProperties;


/**
 * This class generates proxy certificates for use with Globus-based tools.
 * These are "legacy" proxies according to http://dev.globus.org/wiki/Security/ProxyCertTypes <br/>
 * 
 * (Originally from the UNICORE 5 proxy cert plugin. Modified for use with UNICORE 6)
 *
 * @author Thomas Kentemich
 * @author schuller
 */
public class ProxyCertificate  {

	private static final Logger logger = Log.getLogger(Log.SECURITY,ProxyCertificate.class);

	private X509V3CertificateGenerator v3CertGen;
	private PrivateKey privateKey;
	private PublicKey publicKey;

	private boolean limited = false;

	//right now this is always false
	private boolean copyExtensionsIntoProxy = false;

	private X509Certificate proxyCertificate;
	private X509Certificate userCertificate;

	private final long validFrom;
	private final long validTo;
	private final int keyLength;
	private final String signatureAlgName;

	public static final String BEGIN_CERT   ="-----BEGIN CERTIFICATE-----";
	public static final String END_CERT     ="-----END CERTIFICATE-----";
	public static final String BEGIN_PRIVATE="-----BEGIN RSA PRIVATE KEY-----";
	public static final String END_PRIVATE  ="-----END RSA PRIVATE KEY-----";

	public static final String PROXY_JKS_PASSWORD="unicore";
	public static final String PROXY_JKS_ALIAS="mykey";

	/**
	 * property for defining the signature algorithm name <br/>
	 * default for RSA certs: SHA1WITHRSAENCRYPTION <br/>
	 * for DSA certs: SHA256WITHDSA
	 */
	public static final String PROXY_SIGNATURE="unicore.proxy.signature.algorithm";

	static{
		Security.addProvider(new BouncyCastleProvider());
	}	

	private ProxyCertificate(IClientProperties sec, boolean limited,
			long validFrom, long validTo, int keyLength) throws Exception{

		X509Certificate issuerCert = sec.getCertificateChain()[0];
		if (! (issuerCert instanceof X509CertificateObject)) {
			CertificateFactory cf=CertificateFactory.getInstance("X.509", "BC");
			issuerCert=(X509Certificate)cf.generateCertificate(new ByteArrayInputStream(issuerCert.getEncoded()));
		}
		this.userCertificate = issuerCert;
		this.limited = limited;
		this.validFrom = validFrom;
		this.validTo = validTo;
		this.keyLength = keyLength;
		String sigProp=sec.getExtraSettings().getProperty(PROXY_SIGNATURE);
		if(sigProp==null){
			sigProp="SHA1WITHRSAENCRYPTION";
			if(sec.getPrivateKey().getAlgorithm().toLowerCase().contains("dsa")){
				sigProp="SHA256WITHDSA";	
			}
		}
		this.signatureAlgName=sigProp;

		v3CertGen=new X509V3CertificateGenerator();
		KeyPairGenerator kpg=KeyPairGenerator.getInstance(sec.getPrivateKey().getAlgorithm());
		kpg.initialize(keyLength);
		KeyPair keys = kpg.generateKeyPair();
		privateKey = keys.getPrivate();
		publicKey = keys.getPublic();

		this.proxyCertificate = newproxyCertificate(sec.getPrivateKey());
	}

	/**
	 * Returns the proxy certifcate signed with the users private key
	 * 
	 * @return the proxy certifcate
	 */
	public X509Certificate getProxyCertificate() {
		return proxyCertificate;
	}

	/**
	 * Returns the private unprotected key of the proxy certificate
	 * 
	 * @return The private key
	 */
	public PrivateKey getPrivateKey() {
		// key is rubbish anyway, we can give it away
		return privateKey;
	}

	/**
	 * Returns the public key of the proxy certificate 
	 */
	public PublicKey getPublicKey() {
		return publicKey;
	}

	/**
	 * Returns the cert chain (proxy certificate and user certificate) 
	 */
	public Certificate[] getCertChain() {
		return new Certificate[]{getProxyCertificate(),getUserCertificate()};
	}

	/**
	 * Returns start time of the proxy certificate
	 * 
	 * @return Start time
	 */
	public long getValidFrom() {
		return this.validFrom;
	}

	/**
	 * Returns end time of the proxy certificate
	 * 
	 * @return End Time
	 */
	public long getValidTo() {
		return this.validTo;
	}

	/**
	 * Returns the user certificate
	 * 
	 * @return The user certificate
	 */
	public X509Certificate getUserCertificate() {
		return userCertificate;
	}

	/**
	 * Is this a limited proxy?
	 * 
	 * @return Limited or not
	 */
	public boolean isLimited() {
		return limited;
	}

	/**
	 * Return the length of the public/private key pair
	 * 
	 * @return Length
	 */
	public int getKeyLength() {
		return keyLength;
	}

	/**
	 * Generate a new proxy certificate for each call. Subject and issuer is the
	 * current default identity.
	 * 
	 * @param limited - Generate a limited proxy ?
	 * @param lifeTime - proxy lifetime in millis
	 * @param keyLength - key length (one of 512, 1024, 2048, 4096)
	 * @exception Exception
	 */
	public static ProxyCertificate getInstance(IClientProperties sec, 
			boolean limited, long lifeTime, int keyLength) throws Exception {
		logger.debug("Limited: " + limited + " valid: " + lifeTime
				+ " key size: " + keyLength);
		// give 5 minutes grace period here to avoid difficulties with 
		// un-synchronized clocks
		long validFrom = System.currentTimeMillis() - 300000;
		long validTo = System.currentTimeMillis() + lifeTime;
		if (validTo <= validFrom) {
			throw new IllegalArgumentException(
					"Invalid lifetime for proxy certificate");
		}
		if ((keyLength != 512) && (keyLength != 1024) && (keyLength != 2048)
				&& (keyLength != 4096)) {
			throw new IllegalArgumentException(
					"Invalid key length for proxy certificate");
		}
		return new ProxyCertificate(sec, limited, validFrom, validTo,keyLength);
	}


	private X509Certificate newproxyCertificate(PrivateKey issuerKey) throws Exception {

		// expand the users CN with the CN=NNNNN entry
		String proxyExpand = "proxy";
		if (limited) {
			// special tag for limited proxy certificates
			proxyExpand = "limited proxy";
		}

		X509Name issuerDN = (X509Name)userCertificate.getSubjectDN();
		X509NameHelper issuer  = new X509NameHelper(issuerDN);

		X509NameHelper subject = new X509NameHelper(issuerDN); 
		subject.add(X509Name.CN, proxyExpand);

		//
		// create the certificate
		//
		v3CertGen.setIssuerDN(issuer.getAsName());
		v3CertGen.setSubjectDN(subject.getAsName());
		Random rand=new Random();
		v3CertGen.setSerialNumber(new BigInteger(20,rand));
		v3CertGen.setNotBefore(new Date(validFrom));
		v3CertGen.setNotAfter(new Date(validTo));
		// this is the *new* public key !!
		v3CertGen.setPublicKey(publicKey);
		v3CertGen.setSignatureAlgorithm(signatureAlgName);

		if(copyExtensionsIntoProxy){
			Set<String>extensions=userCertificate.getCriticalExtensionOIDs();
			if(extensions!=null){
				for(String ext: extensions){
					byte[] value=userCertificate.getExtensionValue(ext);
					v3CertGen.addExtension(ext, true, getExtensionValue(value));
				}
			}
			extensions=userCertificate.getNonCriticalExtensionOIDs();
			if(extensions!=null){
				for(String ext: extensions){
					byte[] value=userCertificate.getExtensionValue(ext);
					v3CertGen.addExtension(ext, false, getExtensionValue(value));
				}
			}
		}
		// generate cert and sign with the issuer*s private key
		X509Certificate proxyCert = v3CertGen.generate(issuerKey);
		// recheck
		proxyCert.checkValidity(new Date());
		proxyCert.verify(userCertificate.getPublicKey());
		logger.debug("Generated proxyCert for: \n" + proxyCert.getSubjectDN());
		return proxyCert;
	}

	public String getPEMEncoded()throws Exception{
		StringWriter sb=new StringWriter();
		writePEM(sb);
		return sb.toString();
	}

	/**
	 * write the proxy to the given writer in OpenSSL compatible format
	 * 
	 * @param writer
	 * @throws IOException
	 */
	public void writePEM(Writer writer)throws IOException{
		PEMWriter pw=new PEMWriter(writer);
		pw.writeObject(getProxyCertificate());
		pw.writeObject(privateKey);
		pw.writeObject(userCertificate);
		pw.flush();
	}

	/**
	 * write to a JKS keystore (password is "unicore")
	 * @param fileName - the filename of the keystore to write
	 * @throws IOException
	 */
	public void writeJKS(String fileName)throws IOException{
		char[] pass=PROXY_JKS_PASSWORD.toCharArray();
		OutputStream out=null;
		try{
			out=new FileOutputStream(fileName);
			KeyStore ks=KeyStore.getInstance("JKS");
			ks.load(null, pass);
			ks.setKeyEntry(PROXY_JKS_ALIAS, getPrivateKey(), pass, getCertChain());
			ks.store(out, pass);
		}catch(Exception ex){
			IOException ioe=new IOException("Can't write keystore.");
			ioe.initCause(ex);
			throw ioe;
		}finally{
			if(out!=null)out.close();
		}
	}

	//adapted from Java COG Kit
	private static class X509NameHelper { 

		private ASN1Sequence seq;

		/**
		 * Creates an instance using existing {@link X509Name X509Name} 
		 * object. 
		 * This behaves like a copy constructor.
		 *
		 * @param name existing <code>X509Name</code> 
		 */
		public X509NameHelper(X509Name name) {
			this.seq=(ASN1Sequence)name.getDERObject();
		}
		/**
		 * Converts to {@link X509Name X509Name} object.
		 *
		 * @return the <code>X509Name</code> object.
		 */
		public X509Name getAsName() {
			return new X509Name(this.seq);
		}

		/**
		 * Appends the specified OID and value pair
		 * name component to the end of the current
		 * name.
		 * 
		 * @param oid the name component oid, e.g.
		 *         {@link X509Name#CN X509Name.CN}
		 * @param value the value (e.g. "proxy")
		 *
		 */
		public void add(DERObjectIdentifier oid,  String value) {
			ASN1EncodableVector v = new ASN1EncodableVector();
			v.add(oid);
			v.add(new DERPrintableString(value));
			add(new DERSet(new DERSequence(v)));
		}

		/**
		 * Appends the specified name component
		 * entry to the current name.
		 * This can be used to add handle multiple
		 * AVAs in one name component.
		 * 
		 * @param entry the name component to add.
		 */
		public void add(ASN1Set entry) {
			ASN1EncodableVector v = new ASN1EncodableVector();
			int size = seq.size();
			for (int i=0;i<size;i++) {
				v.add(seq.getObjectAt(i));
			}
			v.add(entry);
			seq = new DERSequence(v);
		}

	}
	/**
	 * Retrieves the actual value of the X.509 extension.
	 * 
	 * @param certExtValue the DER-encoded OCTET string value of the extension.
	 * @return the decoded/actual value of the extension (the octets).
	 */
	private byte[] getExtensionValue(byte [] certExtValue) throws IOException {
		ByteArrayInputStream inStream = new ByteArrayInputStream(certExtValue);
		ASN1InputStream derInputStream = new ASN1InputStream(inStream);
		DERObject object = derInputStream.readObject();
		if (object instanceof ASN1OctetString) {
			return ((ASN1OctetString)object).getOctets();
		} else {
			throw new IOException();
		}
	}

}
