/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package de.fzj.unicore.uas.client;

import java.util.Arrays;
import java.util.List;

import javax.xml.namespace.QName;

import org.oasisOpen.docs.wsrf.rp2.GetResourcePropertyDocumentDocument1;
import org.oasisOpen.docs.wsrf.rp2.GetResourcePropertyDocumentResponseDocument;
import org.oasisOpen.docs.wsrf.sg2.AddDocument;
import org.oasisOpen.docs.wsrf.sg2.AddResponseDocument;
import org.oasisOpen.docs.wsrf.sg2.ContentType;
import org.oasisOpen.docs.wsrf.sg2.EntryDocument;
import org.oasisOpen.docs.wsrf.sg2.EntryType;
import org.oasisOpen.docs.wsrf.sg2.ServiceGroupRPDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.Registry;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xfire.ClientException;
import de.fzj.unicore.wsrflite.xfire.WSRFClientFactory;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnavailableFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnknownFault;
import eu.unicore.security.util.client.IClientProperties;

/**
 * Client for accessing a Registry service.
 * 
 * Allows to add entries, and to list services based on porttype and filtering criteria.
 * 
 * To allow service discovery, services have to publish their EPR in a special form
 * containing a metadata item. 
 * 
 * <code>
 * EndpointReferenceType epr= ... your service epr... ;
 * WSServerUtilities.addPortType(... your service porttype ...) ;
 * registryClient.addRegistryEntry(epr);
 * </code>
 * 
 * You can add more content to the registry entry, for example resource properties.
 * 
 * @see de.fzj.unicore.uas.impl.UASWSResourceImpl#getPortType()
 * @see de.fzj.unicore.uas.util.WSServerUtilities#addPortType(EndpointReferenceType, QName)
 * 
 * @author demuth
 * @author schuller
 */
public class LocalRegistryClient{
	
	/**
	 * the QName of the Entry resource property
	 */
	public static final QName ENTRY_RP=EntryDocument.type.getDocumentElementName();
	
	private final EndpointReferenceType epr;
	private final String url;
	
	private Registry reg;
	
	private final Kernel kernel;
	
	public LocalRegistryClient(String resID, Kernel kernel)throws Exception {
		this.kernel=kernel;
		epr=EndpointReferenceType.Factory.newInstance();
		url="xfire.local://"+UAS.REG;
		epr.addNewAddress().setStringValue(url+"?res="+resID);
		WSServerUtilities.addUGSRefparamToEpr(epr,resID);
	} 
	
	public LocalRegistryClient(Kernel kernel)throws Exception {
		this("default_registry", kernel);
	} 
	
	protected synchronized Registry getRegistry()throws ClientException{
		if(reg==null){
			IClientProperties sec=kernel.getSecurityProperties();
			try{
			     reg=new WSRFClientFactory(sec).createProxy(Registry.class, url, epr);
			}
			catch(Exception ex){
				throw new ClientException(ex);
			}
		}
		return reg;
	}
	
	/**
	 * add a registry entry
	 * @param in
	 * @return
	 * @throws Exception
	 */
	public AddResponseDocument addRegistryEntry(AddDocument in) throws Exception{
		return getRegistry().Add(in);
	}

	/**
	 * add an entry to the registry configured for UAS
	 * 
	 * @param memberEpr the epr of the service
	 * @param content The content document (ResourceProperty doc)
	 * @throws Exception
	 * @return the epr of the new sg entry
	 */
	public AddResponseDocument addRegistryEntry(EndpointReferenceType memberEpr,
			ContentType content) throws Exception{
		AddDocument in=AddDocument.Factory.newInstance();
		in.addNewAdd().setContent(content);
		in.getAdd().setMemberEPR(memberEpr);
		return addRegistryEntry(in);
	}
	
	/**
	 * add an entry to the registry (with no content)
	 * 
	 * @param memberEpr
	 * @return
	 * @throws Exception
	 */
	public AddResponseDocument addRegistryEntry(EndpointReferenceType memberEpr) throws Exception{
		ContentType content=ContentType.Factory.newInstance();
		content.setNil();
		return addRegistryEntry(memberEpr,content);
	}
	
	
	/**
	 * lists all the entries in this local registry
	 * @return
	 * @throws Exception
	 */
	public List<EntryType> listEntries()throws Exception{
		EntryType[] entries=getResourcePropertiesDocument().getServiceGroupRP().getEntryArray();
		return Arrays.asList(entries);
	}
	
	/**
	 * returns the local registry's resource properties document
	 */
	ServiceGroupRPDocument getResourcePropertiesDocument()throws Exception{
		GetResourcePropertyDocumentResponseDocument grprd=GetResourcePropertyDocumentResponseDocument.
		Factory.parse(getResourcePropertyDocument());
		ServiceGroupRPDocument doc=	ServiceGroupRPDocument.Factory.
		parse(grprd.getGetResourcePropertyDocumentResponse().newInputStream());
		return doc;
	}
	
	String getResourcePropertyDocument() throws BaseFault,ResourceUnknownFault,ResourceUnavailableFault,ClientException{
		GetResourcePropertyDocumentDocument1 req=GetResourcePropertyDocumentDocument1.Factory.newInstance();
		req.addNewGetResourcePropertyDocument();
		return String.valueOf(getRegistry().GetResourcePropertyDocument(req));
	}
		
}
