package de.fzj.unicore.uas.fts.rft;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import de.fzj.unicore.uas.client.ReliableFileTransferClient;
import de.fzj.unicore.uas.client.ReliableFileTransferClient.Chunk;
import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.io.XnjsFile;
import de.fzj.unicore.xnjs.tsi.TSI;
import de.fzj.unicore.xnjs.util.AsyncCommandHelper;
import eu.unicore.security.Client;

public class StoreImpl implements ReliableFileTransferClient.Store{

	final TSI storage;
	final long totalLength;
	final String target;
	final String partsDir;
	final String propertiesPath;
	final Properties properties=new Properties();
	final Client client;
	final Configuration config;
	final int chunkLength;
	
	public static int minChunkLength=10*1024*1024;
	
	//the chunks directory name is built from the original filename and this suffix
	private static final String suffix=".unicore_rft.parts";

	public StoreImpl(Client client,TSI storage, Configuration config, String target, long totalLength){
		this.storage=storage;
		this.totalLength=totalLength;
		this.target=target;
		this.partsDir=target+suffix;
		this.client=client;
		this.config=config;
		this.propertiesPath=partsDir+"/properties";
		this.chunkLength=computeChunkLength(totalLength);
	}

	/**
	 * compute the chunk length according to the following rules<br/>
	 * <ul>
	 *  <li>not more than 100 chunks are created</li>
	 *  <li>the minumum size of a chunk is {@link #minChunkLength}</li>
	 * </ul>
	 * @param totalLength
	 * @return
	 */
	public static int computeChunkLength(long totalLength){
		int length=minChunkLength;
		int total=minChunkLength*100;
		if(total<totalLength){
			length+=(totalLength-total)/99;
		}
		return length;
	}
	
	@Override
	public List<Chunk> getChunks() throws Exception{
		List<Chunk>result=allChunks();
		Iterator<Chunk>iterator=result.iterator();
		while(iterator.hasNext()){
			if(!needToDownload(iterator.next()))iterator.remove();
		}
		return result;
	}

	@Override
	public void ack(Chunk chunk) throws Exception {
		ChunkImpl c=(ChunkImpl)chunk;
		properties.setProperty("chunk."+format(c.getIndex())+".status","done");
		writeProperties();
	}

	public void finish()throws Exception{
		if(properties.getProperty("merge.complete")==null){
			String cmd="/bin/cat "+buildFileList();
			AsyncCommandHelper ach=new AsyncCommandHelper(config, cmd, "merge_files", null, client, storage.getStorageRoot()+"/"+partsDir+"/");
			ach.setStdout("../../"+target);
			ach.submit();
			while(!ach.isDone()){
				Thread.sleep(2000);
			}
			if(!ach.getResult().getResult().isSuccessful()){
				throw new Exception("Merge was not successful.");
			}
		}
		properties.setProperty("merge.complete","true");
		writeProperties();
		storage.rmdir(partsDir);
	}
	
	String buildFileList(){
		StringBuilder sb=new StringBuilder();
		int i=0;
		long pos=0;
		while(pos<totalLength){
			if(pos>0)sb.append(" ");
			sb.append("p_"+format(i));
			pos+=chunkLength;
			i++;
		}
		return sb.toString();
	}
	
	//check if this chunk has to be (re-)downloaded
	public boolean needToDownload(Chunk chunk){
		ChunkImpl c=(ChunkImpl)chunk;
		String status=properties.getProperty("chunk."+format(c.getIndex())+".status",null);
		return (!"done".equals(status));
	}
	
	protected String format(int index){
		return String.format("%04d", index);
	}
	
	protected void init()throws Exception{
		XnjsFile parts=storage.getProperties(target+suffix);
		if(parts==null){
			createNewParts();
		}
		initParts();
	}
	
	protected List<Chunk> allChunks()throws Exception{
		init();
		List<Chunk>result=new ArrayList<Chunk>();
		int i=0;
		long pos=0;
		while(pos<totalLength){
			long length=Math.min(chunkLength, totalLength-pos);
			String path=partsDir+"/"+"p_"+format(i);
			ChunkImpl c=new ChunkImpl(i,storage, path, pos, length-1);
			result.add(c);
			pos+=chunkLength;
			i++;
		}
		return result;
	}
	
	protected void createNewParts()throws Exception{
		storage.mkdir(partsDir);
	}
	
	protected void initParts()throws Exception{
		XnjsFile propFile=storage.getProperties(propertiesPath);
		if(propFile!=null){
			InputStream is=storage.getInputStream(propertiesPath);
			try{
				properties.load(is);
			}finally{
				is.close();
			}
		}
	}
	
	protected void writeProperties()throws Exception{
		properties.store(storage.getOutputStream(propertiesPath), "auto-generated, DO NOT EDIT");
	}
}
