/*********************************************************************************
 * Copyright (c) 2006-2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 
package de.fzj.unicore.uas.impl.registry;

import java.util.Calendar;

import org.apache.log4j.Logger;
import org.oasisOpen.docs.wsrf.rl2.DestroyDocument;
import org.oasisOpen.docs.wsrf.rl2.DestroyResponseDocument;
import org.oasisOpen.docs.wsrf.rl2.SetTerminationTimeDocument;
import org.oasisOpen.docs.wsrf.rl2.SetTerminationTimeResponseDocument;
import org.oasisOpen.docs.wsrf.sg2.AddDocument;
import org.oasisOpen.docs.wsrf.sg2.AddResponseDocument;

import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.client.ExternalRegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceNotDestroyedFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnavailableFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnknownFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.TerminationTimeChangeRejectedFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.UnableToSetTerminationTimeFault;
import eu.unicore.security.util.Log;


/**
 * Local Registry. <br/>
 * Entries in the local registry can be pushed to one or more external registries,
 * with a simple update/refresh mechanism. The lifetime of a local registry entry is determined 
 * by the smallest termination time returned by the external registries minus a grace period (60 secs).
 * In case the external registries fail to respond, the lifetime supplied to Add() is used, or 
 * as fallback, a fixed interval of 5 minutes. This mechanism ensures that the external registries are 
 * in a reasonably up-to-date state.
 *  
 * 
 * @author demuth
 * @author schuller
 */
public class LocalRegistryImpl extends RegistryImpl {

	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES,LocalRegistryImpl.class);
	
	protected static final long readd_offset = 60000; 
	
	/**
	 * Add a service to this service group. This will create a new ws resource to represent the entry.
	 * The lifetime of the new entry is determined by the external registry, or 
	 * the time supplied to this method, or a 5 minute fallback.
	 */
	public AddResponseDocument Add(AddDocument in) throws BaseFault {
		logger.debug("Adding entry to local registry.");
		
		
		Calendar terminationTime=null;
		Object o=in.getAdd().getInitialTerminationTime();
		if(o!=null){
			terminationTime=makeCalendar(o);
		}
		else{
			terminationTime=Calendar.getInstance();
			terminationTime.add(Calendar.MINUTE, 5);			
		}
		
		RegistryHandler rh=null;
		try{
			rh=RegistryHandler.get(kernel);
		}catch(Exception ex){
			Log.logException("Can't setup registry", ex);
			throw BaseFault.createFault("Internal server error", ex, true);
		}
		
		if(rh.usesExternalRegistry()){
			try {
				logger.debug("Trying to add entry to external registry.");
				ExternalRegistryClient externalRegistryClient = rh.getExternalRegistryClient();
				terminationTime= externalRegistryClient.addRegistryEntry(in);
				if(logger.isDebugEnabled()){
					logger.debug("Will try to re-add entry for "
							+ in.getAdd().getMemberEPR().getAddress().getStringValue() 
							+ " at: "+terminationTime.getTime().toString());
				}
			} catch (Exception e) {
				LogUtil.logException("Could not connect to external Registry!",e,logger);
			}
			in.getAdd().setInitialTerminationTime(terminationTime);
		}

		return super.Add(in);
	}
	
	@Override
	public DestroyResponseDocument Destroy(DestroyDocument in) throws ResourceNotDestroyedFault, ResourceUnknownFault, ResourceUnavailableFault {
		throw ResourceNotDestroyedFault.createFault("Not destroyed."); 
	}

	@Override
	public SetTerminationTimeResponseDocument SetTerminationTime(
			SetTerminationTimeDocument in)
			throws UnableToSetTerminationTimeFault,
			TerminationTimeChangeRejectedFault, ResourceUnknownFault,
			ResourceUnavailableFault {
		throw TerminationTimeChangeRejectedFault.createFault("Not changed.");
	}
}
