/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 
package de.fzj.unicore.uas.impl.registry;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;



/**
 * @author Henning Mersch <h.mersch@fz-juelich.de
 * @author Bastian Demuth <b.demuth@fz-juelich.de>
 */
public class MulticastRegistryServer extends Thread {
	
	private static final Logger logger=LogUtil.getLogger(LogUtil.UNICORE,MulticastRegistryServer.class);
	
	private final InetAddress bgroup;
	private final int port;
	private final Set<String> keys=new HashSet<String>();
	private final String server;
	public static final String MULTICAST_KEY = "UNICORE6Registry";
	private volatile boolean running=false;
	
	/**
	 * Creates a new instance of MulticastRegistryAdvertiser
	 * @param bgroup - the broadcast group to use
	 * @param port - the listen port
	 * @param server - the server address to advertise
	 * @param keys - a list of Strings used as keys to react on
	 */
	public MulticastRegistryServer(InetAddress bgroup, int port, String server, String... keys) {
		this.bgroup = bgroup;
		this.port = port;
		this.keys.addAll(Arrays.asList(keys));
		this.server = server;
	}
	
	/**
	 * Creates a new instance of MulticastRegistryServer with default values
	 * <ul>
	 * <li>bgroup: 228.5.6.7
	 * <li>port: 7700
	 * <li>key: UNICORE6Registry
	 * <li>server: local registry URL
	 * </ul>
	 */
	public MulticastRegistryServer(Kernel kernel)throws UnknownHostException{
		this(InetAddress.getByName("228.5.6.7"), 
				7700, MULTICAST_KEY,
				WSServerUtilities.makeAddress(UAS.REG, "default_registry",kernel));
	}

	public void run() {
		logger.info("Starting MulticastRegistryServer.");
		try {
			MulticastSocket listenSocket = new MulticastSocket(port);
			listenSocket.joinGroup(bgroup);
			listenSocket.setSoTimeout(2000); // timout: .5 sec.
			byte[] buf = new byte[1000];
			DatagramPacket recv = new DatagramPacket(buf, buf.length);
			while(!isInterrupted()) { //check every 2000 msec if interruypted
				try {
					running=true;
					listenSocket.receive(recv);
					String request = new String(recv.getData()).trim();
					String requestKey = request.trim().split(":")[1];
					logger.debug("Got request: "+request+ " from "+recv.getAddress()+" key="+requestKey);
					if(keys.contains(requestKey)) {
						Thread.sleep(500); // wait a bit to give to client change to get to answermode
						sendResponse("Response:"+MULTICAST_KEY+":"+server);
					} else {
						logger.debug("Ignoring unknown request:"+request);
					}
				} catch (IOException e) {
					//logger.debug(e);
				}
			}
			running=false;
			
		} catch (Exception ex) {
			logger.warn("Error",ex);
		}
	}
	
	public void sendResponse(String response) throws IOException {
		//send response
		MulticastSocket sendSocket = new MulticastSocket(port);

		sendSocket.joinGroup(bgroup);
		DatagramPacket hi = new DatagramPacket(response.getBytes(),
				response.length(),
				bgroup, port);
		sendSocket.send(hi);
		logger.info("Sent response back: "+response);
	}
	
	public boolean isRunning(){
		return running;
	}
}


