package de.fzj.unicore.uas.impl.sms;

import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.FileSystemType;
import org.oasisOpen.docs.wsrf.rl2.DestroyDocument;
import org.oasisOpen.docs.wsrf.rl2.DestroyResponseDocument;
import org.oasisOpen.docs.wsrf.rl2.SetTerminationTimeDocument;
import org.oasisOpen.docs.wsrf.rl2.SetTerminationTimeResponseDocument;
import org.oasisOpen.docs.wsrf.rl2.TerminationTimeDocument.TerminationTime;
import org.unigrids.services.atomic.types.PropertyType;
import org.unigrids.x2006.x04.services.smf.AccessibleStorageEnumerationDocument;
import org.unigrids.x2006.x04.services.smf.CreateSMSDocument;
import org.unigrids.x2006.x04.services.smf.CreateSMSResponseDocument;
import org.unigrids.x2006.x04.services.smf.StorageBackendParametersDocument.StorageBackendParameters;
import org.unigrids.x2006.x04.services.smf.StorageDescriptionType;
import org.unigrids.x2006.x04.services.smf.StorageEnumerationDocument;
import org.unigrids.x2006.x04.services.smf.StorageFactoryPropertiesDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.Enumeration;
import de.fzj.unicore.uas.StorageFactory;
import de.fzj.unicore.uas.StorageManagement;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.impl.UASWSResourceImpl;
import de.fzj.unicore.uas.impl.dsms.DSMSImpl;
import de.fzj.unicore.uas.impl.enumeration.EnumerationImpl;
import de.fzj.unicore.uas.impl.sms.StorageManagementHomeImpl.StorageTypes;
import de.fzj.unicore.uas.util.AddOnStorageDescriptionFactory;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.exceptions.ResourceUnknownException;
import de.fzj.unicore.wsrflite.messaging.PullPoint;
import de.fzj.unicore.wsrflite.persistence.Persist;
import de.fzj.unicore.wsrflite.utils.PropertyHelper;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceNotDestroyedFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnavailableFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnknownFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.TerminationTimeChangeRejectedFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.UnableToSetTerminationTimeFault;
import de.fzj.unicore.wsrflite.xmlbeans.rp.ImmutableResourceProperty;

/**
 * Implements the storage factory
 * 
 * @author schuller
 * @author daivandy
 * @since 6.3
 */
public class StorageFactoryImpl extends UASWSResourceImpl implements StorageFactory{

	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES, StorageFactoryImpl.class);

	//properties for configuring the storage factory, similar to the
	//corresponding targetsystem properties
	public static final String PREFIX="uas.storagefactory";
	public static final String TYPE=PREFIX+".type";
	public static final String PATH=PREFIX+".path";
	public static final String CLASS=PREFIX+".class";
	public static final String CLEANUP=PREFIX+".cleanup";

	/**
	 * maps SMS to SMS owner DNs 
	 */
	@Persist
	private Map<String, String>smsOwners=new HashMap<String, String>();

	/**
	 * get the owner for a given sms
	 * @param smsID
	 * @return owner DN
	 */
	public String getOwnerForSMS(String smsID){
		return smsOwners.get(smsID);
	}

	public SMSReferenceResourceProperty getSMSReferenceProperty(){
		return (SMSReferenceResourceProperty)properties.get(RPSMSReferences);
	}

	@Override
	public void initialise(String serviceName, Map<String, Object> initParams)
	throws Exception {
		super.initialise(serviceName, initParams);
		properties.put(RPSMSReferences, new SMSReferenceResourceProperty(new EndpointReferenceType[0]));
		properties.put(RPAccessibleSMSReferences, new AccessibleSMSReferenceRP(this));
		try{
			StorageEnumerationDocument sred=StorageEnumerationDocument.Factory.newInstance();
			sred.setStorageEnumeration(createEnumeration(RPSMSReferences));
			properties.put(RPSMSEnumeration, new ImmutableResourceProperty(sred));
		}
		catch(Exception ex){
			LogUtil.logException("Error creating SMS reference enumeration",ex,logger);
		}
		try{
			AccessibleStorageEnumerationDocument sred=AccessibleStorageEnumerationDocument.Factory.newInstance();
			sred.setAccessibleStorageEnumeration(createEnumeration(RPAccessibleSMSReferences));
			properties.put(RPAccessibleSMSEnumeration, new ImmutableResourceProperty(sred));
		}
		catch(Exception ex){
			LogUtil.logException("Error creating SMS reference enumeration",ex,logger);
		}

		String xnjsRef=null; //TODO: how to handle non-default XNJS instances
		properties.put(RPStorageDescription,new StorageDescriptionRP(this,xnjsRef));
		logger.info("Storage factory <"+getUniqueID()+"> created");
		publish();
	}

	public CreateSMSResponseDocument CreateSMS(CreateSMSDocument in)
	throws BaseFault {
		if(logger.isTraceEnabled())logger.trace("CreateSMS: "+in);
		
		String clientName=(getClient()!=null?getClient().getDistinguishedName():"<no client>");
		try{
			String storageBackendType = "DEFAULT";
			
			String cleanup = "true";
			String storageName = "Dynamic";
			
			//extra settings for the to-be-created SMS
			Map<String,String>extraProperties=new HashMap<String,String>();
			
			// accommodate for parametrisable storage backend type
			StorageDescriptionType sdType = in.getCreateSMS().getStorageDescription();
			if(sdType!=null) {
				if(sdType.getStorageBackendType()!=null){
					storageBackendType = sdType.getStorageBackendType();
				}
				FileSystemType fsType = sdType.getFileSystem();
				if(fsType!=null) storageName = fsType.getName();
				if(sdType.getStorageBackendParameters()!=null){
					extraProperties.putAll(getAdditionClientConfigurationItems(sdType.getStorageBackendParameters()));
				}
			}

			//we parse properties starting with this prefix
			String typePrefix = PREFIX+"."+storageBackendType+".";
			
			String storageTypeP=kernel.getProperty(typePrefix+"type",StorageTypes.FIXEDPATH.toString());
			Map<String, String> typeProperties = new PropertyHelper(kernel.getProperties(), typePrefix).getFilteredMap();
			
			extraProperties.putAll(getAdditionalConfigurationItems(typeProperties, storageBackendType));
			
			String workdir = typeProperties.get(typePrefix+"path");
			if(workdir==null && StorageTypes.FIXEDPATH.equals(storageTypeP)){
				//check old property as well
				workdir=kernel.getProperty("uas.storagefactory.path");
				if(workdir!=null){
					logger.warn("Configuration uses deprecated property <uas.storagefactory.path>. Should use <"+
							typePrefix+"path> instead");
				}
			}
			String storageClass = typeProperties.get(typePrefix+"class");
			if(storageClass==null)storageClass=FixedStorageImpl.class.getName();
			cleanup = typeProperties.get(typePrefix+"cleanup");
			if(cleanup==null || cleanup.equalsIgnoreCase("")) cleanup="true";
		
			StorageTypes storageType=StorageTypes.valueOf(storageTypeP);
			
			Map<String,Object>initMap=new HashMap<String,Object>();

			initMap.put(SMSBaseImpl.INIT_DIRECTORY,workdir);
			initMap.put(SMSBaseImpl.INIT_DIRECTORY_APPEND_UNIQUE_ID,Boolean.TRUE);
			initMap.put(StorageManagementHomeImpl.INIT_STORAGECLASS,storageClass);
			initMap.put(SMSBaseImpl.INIT_ADDITIONAL_SETTINGS, extraProperties);

			TerminationTime tt=in.getCreateSMS().getTerminationTime();
			if(tt!=null){
				initMap.put(SMSBaseImpl.INIT_INITIAL_TERMINATION_TIME,tt.getCalendarValue());
			}
			
			initMap.put(SMSBaseImpl.INIT_CLEANUP_ON_DESTROY,Boolean.parseBoolean(cleanup));
			initMap.put(SMSBaseImpl.INIT_FACTORYID, getUniqueID());
			initMap.put(SMSBaseImpl.INIT_STORAGENAME, storageName);
			initMap.put(SMSBaseImpl.INIT_PROTOCOLS_PROPERTY, typePrefix+"protocols");
			initMap.put(StorageManagementHomeImpl.INIT_STORAGETYPE,storageType);

			EndpointReferenceType epr=createStorageResource(initMap);

			CreateSMSResponseDocument resD=CreateSMSResponseDocument.Factory.newInstance();
			resD.addNewCreateSMSResponse().setSmsReference(epr);
			
			String smsID=WSServerUtilities.extractResourceID(epr);
			initStorage(smsID);
			logger.info("Created new StorageManagement resource <"+smsID+"> for "+clientName);
			smsOwners.put(smsID, clientName);
			getSMSReferenceProperty().add(epr);
			setDirty();
			return resD;
		}catch(Exception ex){
			LogUtil.logException("Could not create Storage instance for <"+clientName+">", ex, logger);
			throw BaseFault.createFault("Could not create Storage instance. Please consult the site administrator.", ex);
		}
	}

	/**
	 * replaces prefix of the storage factory into a normal one of SMS
	 * @param original
	 * @return
	 */
	private Map<String, String> getAdditionalConfigurationItems(Map<String, String> original, 
			String storageBackendType) {
		Map<String, String> ret = new HashMap<String, String>();
		int len = (PREFIX + storageBackendType + ".").length() + 1;
		for (String key: original.keySet()) {
			String newKey = AddOnStorageDescriptionFactory.PREFIX +  
				key.substring(len);
			ret.put(newKey, original.get(key));
		}
		return ret;
	}
	/**
	 * replaces prefix of the storage factory into a normal one of SMS
	 * @param params - parameters provided by the client
	 * @return
	 */
	private Map<String, String> getAdditionClientConfigurationItems(StorageBackendParameters params) {
		Map<String, String> ret = new HashMap<String, String>();
		for(PropertyType p: params.getPropertyArray()){
			ret.put(p.getName(), p.getValue());
		}
		return ret;
	}
	
	protected EndpointReferenceType createEnumeration(QName rp)throws Exception{
		Map<String,Object>init=new HashMap<String, Object>();
		init.put(EnumerationImpl.INIT_TARGETSERVICE_EPR, this.getEPR());
		init.put(EnumerationImpl.INIT_TARGETSERVICE_RP, rp);
		Calendar c=Calendar.getInstance();
		c.add(Calendar.MONTH, 24);
		init.put(EnumerationImpl.INIT_INITIAL_TERMINATION_TIME, c);
		Home h=kernel.getHome(UAS.ENUMERATION);
		if(h==null)throw new Exception("Enumeration service is not deployed!");
		String id=h.createWSRFServiceInstance(init);
		return WSServerUtilities.makeEPR(UAS.ENUMERATION,id,Enumeration.ENUM_PORT, kernel);
	}


	/**
	 * create new storage management service and return its epr
	 * @param initialisation params
	 * @return EPR or null if SMS could not be created
	 */
	protected EndpointReferenceType createStorageResource(Map<String,Object> initParam)
	throws Exception{
		Home home=kernel.getHome(UAS.SMS);
		if(home==null){
			throw new ResourceUnknownException("Storage management service is not deployed.");
		}
		String id=home.createWSRFServiceInstance(initParam);
		return WSServerUtilities.makeEPR(UAS.SMS,id,StorageManagement.SMS_PORT, kernel);
	}

	protected void initStorage(String uniqueID)throws Exception{
		StorageManagementHomeImpl smsHome=(StorageManagementHomeImpl)kernel.getHome(UAS.SMS);
		SMSBaseImpl sms=(SMSBaseImpl)smsHome.get(uniqueID);
		if(!(sms instanceof DSMSImpl)){
			sms.getStorageAdapter().mkdir("/");
		}
	}

	@Override
	public QName getPortType() {
		return SMF_PORT;
	}

	@Override
	public QName getResourcePropertyDocumentQName() {
		return StorageFactoryPropertiesDocument.type.getDocumentElementName();
	}

	@Override
	public DestroyResponseDocument Destroy(DestroyDocument in) throws ResourceNotDestroyedFault, ResourceUnknownFault, ResourceUnavailableFault {
		throw ResourceNotDestroyedFault.createFault("Not destroyed."); 
	}

	@Override
	public SetTerminationTimeResponseDocument SetTerminationTime(
			SetTerminationTimeDocument in)
	throws UnableToSetTerminationTimeFault,
	TerminationTimeChangeRejectedFault, ResourceUnknownFault,
	ResourceUnavailableFault {
		throw TerminationTimeChangeRejectedFault.createFault("Not changed.");
	}


	@Override
	public void customPostActivate(){
		//check for deleted SMSs and remove them...
		try{
			PullPoint p=kernel.getMessaging().getPullPoint(getUniqueID());
			SMSReferenceResourceProperty smsRef=getSMSReferenceProperty();
			while(p.hasNext()){
				String m=(String)p.next().getBody();
				if(m.startsWith("deleted:")){
					String id=m.substring(m.indexOf(":")+1);
					logger.debug("Removing Storage with ID "+id+"...");
					smsRef.remove(id);
					smsOwners.remove(id);
					setDirty();
				}
			}
		}catch(Exception e){
			LogUtil.logException(e.getMessage(),e,logger);
		}
	}

}
