/*
 * Copyright (c) 2010 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licensing information.
 *
 * Author: T. Rękawek <newton@mat.uni.torun.pl>
 */

package de.fzj.unicore.uas.dsms;

import static org.junit.Assert.fail;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FileUtils;
import org.unigrids.services.atomic.types.GridFileType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import pl.edu.icm.dsms.catalogue.Catalogue;
import pl.edu.icm.dsms.catalogue.CatalogueClient;
import pl.edu.icm.dsms.catalogue.model.Lda;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.client.BaseUASClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.StorageFactoryClient;
import de.fzj.unicore.uas.impl.UASWSResourceImpl;
import de.fzj.unicore.uas.impl.registry.RegistryHandler;
import de.fzj.unicore.uas.impl.sms.SMSBaseImpl;
import de.fzj.unicore.uas.impl.sms.StorageFactoryHomeImpl;
import de.fzj.unicore.uas.impl.sms.StorageManagementHomeImpl;
import de.fzj.unicore.uas.impl.sms.StorageManagementHomeImpl.StorageTypes;
import de.fzj.unicore.uas.util.CreateSMSOnStartup;
import de.fzj.unicore.uas.xnjs.XNJSFacade;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.exceptions.ResourceNotCreatedException;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.impl.WSResourceImpl;



/**
 * Some utils helpful for testing catalogue.
 * @author newton
 *
 */
public abstract class AbstractTestBase 
{
	private static final String DEFAULT_CAT_NAME = "default_catalogue";
	protected static final String invalidDSmsId = "123";

	private Home factoryHome, smsHome, catalogueHome;
	private String smsResId, factoryResId, catalogueResId;

	protected static Kernel kernel;
	
	protected static void startKernel() throws Exception
	{
		FileUtils.deleteQuietly(new File("target", "data"));
		System.setProperty("java.util.logging.config.file", "src/test/resources/dcatalogue/log4j.properties");
		kernel=new Kernel();
		kernel.startSynchronous("src/test/resources/dcatalogue/uas.config");
		RegistryHandler.get(kernel);
		//start XNJS and link kernel instance to XNJS instance
		XNJSFacade.get(null, kernel);
	}
	
	protected static void shutdownKernel()
	{
		//kernel.shutdown();
	}

	public static void cleanStorageDirs() {
		File testsRoot = new File("target", "unicorex-test");
		File s1 = new File(testsRoot, "storagefactory");
		File s2 = new File(testsRoot, "teststorage");
		FileUtils.deleteQuietly(s1);
		FileUtils.deleteQuietly(s2);
		s1.mkdir();
		s2.mkdir();
	}
	
	protected EndpointReferenceType findFirstAccessibleService(List<EndpointReferenceType>eprs)
	{
		for(EndpointReferenceType epr: eprs)
		{
			try
			{
				BaseUASClient c = new BaseUASClient(epr, kernel.getSecurityProperties());
				c.getCurrentTime();
				return epr;
			}
			catch(Exception e)
			{
			}
		}
		return null;
	}
	
	protected boolean compareEprList(List<EndpointReferenceType> list1, List<Lda> list2)
	{
		if(list1.size() != list2.size())
			return false;
		
		for(EndpointReferenceType epr1 : list1)
		{
			boolean ok = false;
			for(Lda lda : list2)
				if(compareEpr(epr1, lda.getSms()))
				{
					ok = true;
					break;
				}
			if(!ok)
				return false;
		}
		return true;
	}
	
	protected boolean compareFileList(GridFileType[] list1, String[] list2)
	{
		if(list1.length != list2.length)
			return false;
		
		for(GridFileType file1 : list1)
		{
			boolean ok = false;
			for(String file2 : list2)
				if(new File(file1.getPath()).getName().equals(file2))
				{
					ok = true;
					break;
				}
			if(!ok)
				return false;
		}
		return true;

	}
	
	protected boolean compareEpr(EndpointReferenceType sms1, EndpointReferenceType sms2)
	{
		return sms1.getAddress().getStringValue().equals(sms2.getAddress().getStringValue());
	}
		
	protected RegistryClient getRegistryClient() throws Exception
	{
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + UAS.REG + "?res=default_registry");
		
		return new RegistryClient(epr, kernel.getSecurityProperties());
	}

	protected CatalogueClient getCatalogueClient() throws Exception
	{
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + Catalogue.CATALOGUE + "?res=" + DEFAULT_CAT_NAME);
		
		return new CatalogueClient(epr, kernel.getSecurityProperties());
	}
		
	protected StorageClient getSmsClient() throws Exception
	{
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + UAS.SMS + "?res=" + CreateSMSOnStartup.INSTANCE_ID);
		
		return new StorageClient(epr, kernel.getSecurityProperties());
	}

	protected StorageFactoryClient getFactoryClient() throws Exception
	{
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + UAS.SMF + "?res=" + StorageFactoryHomeImpl.DEFAULT_SMF_NAME);
		
		return new StorageFactoryClient(epr, kernel.getSecurityProperties());
	}
	
	protected void createSms() throws ResourceNotCreatedException
	{
		Map<String,Object>map = new HashMap<String,Object>();
		map.put(WSResourceImpl.INIT_UNIQUE_ID, CreateSMSOnStartup.INSTANCE_ID);
		map.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());
		map.put(UASWSResourceImpl.INITPARAM_PUBLISH_TO_REGISTRY, Boolean.TRUE);
		map.put(StorageManagementHomeImpl.INIT_STORAGETYPE, StorageTypes.VARIABLE);
		map.put(SMSBaseImpl.INIT_FILTERLISTINGS, Boolean.TRUE);
		map.put(SMSBaseImpl.INIT_STORAGENAME, "SHARE");
		
		String dir = kernel.getProperty(CreateSMSOnStartup.CONF_WORKDIR);
		map.put(SMSBaseImpl.INIT_DIRECTORY, dir);
		
		smsHome = kernel.getHome(UAS.SMS);
		smsResId = smsHome.createWSRFServiceInstance(map);
	}
	
	protected void destroySms() throws Exception
	{
		smsHome.destroyWSRFServiceInstance(smsResId);
	}
	
	protected void createFactory() throws ResourceNotCreatedException
	{
		String defaultSmfName = StorageFactoryHomeImpl.DEFAULT_SMF_NAME;

		Map<String, Object>imap = new HashMap<String,Object>();
		imap.put(WSResourceImpl.INIT_UNIQUE_ID, defaultSmfName);
		imap.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());
		
		factoryHome = kernel.getHome(UAS.SMF);
		factoryResId = factoryHome.createWSRFServiceInstance(imap);
	}
	
	protected void destroyFactory() throws Exception
	{
		factoryHome.destroyWSRFServiceInstance(factoryResId);
	}
	
	protected void createCatalogue() throws ResourceNotCreatedException
	{		
		Map<String,Object> imap = new HashMap<String,Object>();
		imap.put(WSResourceImpl.INIT_UNIQUE_ID, DEFAULT_CAT_NAME);
		imap.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());
		
		catalogueHome = kernel.getHome(Catalogue.CATALOGUE);
		catalogueResId = catalogueHome.createWSRFServiceInstance(imap);
	}
	
	protected void destroyCatalogue() throws Exception
	{
		catalogueHome.destroyWSRFServiceInstance(catalogueResId);
	}
	
	public static Calendar lifeTime()
	{
		Calendar c=new GregorianCalendar();
		c.add(Calendar.MINUTE, 5);
		return c;
	}
	
	protected static void assertCatEx(CatalogueClient cat, String method, Object[] params,
			Class<?> expectedException)
	{
		Class<?>[] paramTypes = new Class[params.length];
		try
		{
			for(int i=0;i<params.length;i++)
				if(params[i] == null || params[i] instanceof String)
					paramTypes[i] = String.class;
				else if(params[i] instanceof EndpointReferenceType)
					paramTypes[i] = EndpointReferenceType.class;
				else
					paramTypes[i] = params[i].getClass();
			Method m = CatalogueClient.class.getMethod(method, paramTypes);
			m.invoke(cat, params);
		} catch (NoSuchMethodException e)
		{
			StringBuilder builder = new StringBuilder("No such method: ");
			builder.append(method).append('(');
			for(int i=0;i<paramTypes.length;i++)
			{
				builder.append(paramTypes[i].getSimpleName());
				if(i < paramTypes.length-1)
					builder.append(", ");
			}
			builder.append(")");
			e.printStackTrace();
			fail(builder.toString());
		} catch (IllegalArgumentException e)
		{
			e.printStackTrace();
			fail(e.toString());
		} catch (IllegalAccessException e)
		{
			e.printStackTrace();
			fail(e.toString());
		} catch (InvocationTargetException e)
		{
			if(e.getCause().getClass().equals(expectedException))
					return;
		}
		fail(expectedException.toString() + " should have been thrown.");
	}
}