package de.fzj.unicore.uas.client;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;
import org.oasisOpen.docs.wsrf.rp2.UpdateResourcePropertiesDocument;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.fts.TransferredBytesDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.fts.udt.UDTFileTransfer;
import de.fzj.unicore.uas.fts.udt.UDTSession;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.xfire.ClientException;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceNotDestroyedFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnavailableFault;
import de.fzj.unicore.wsrflite.xmlbeans.exceptions.ResourceUnknownFault;
import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.ems.ExecutionException;
import eu.unicore.security.Client;
import eu.unicore.security.util.client.IClientProperties;
import eu.unicore.unicore6.filetransfer.udt.ExchangeParametersRequestDocument;
import eu.unicore.unicore6.filetransfer.udt.ExchangeParametersResponseDocument;
import eu.unicore.unicore6.filetransfer.udt.HostDocument;
import eu.unicore.unicore6.filetransfer.udt.IsSenderDocument;

/**
 * manage file transfer using a direct UDT connection
 * 
 * @author oistrez
 * @author schuller
 * 
 * TODO needs to be reviewed
 */
public class UDTClient extends FileTransferClient {

	private static final Logger logger=LogUtil.getLogger(LogUtil.CLIENT,UDTClient.class);

	private final UDTFileTransfer fts;
	
	private final UDTSession session;
	
	//size of the transferred file
	private long datasize=-1;
	
	public UDTClient(String url, EndpointReferenceType epr, IClientProperties sec) throws Exception{
		super(url, epr, sec);
		fts=makeProxy(UDTFileTransfer.class);
		session=new UDTSession();
	}
	
	public void setConfiguration(Configuration config){
		session.setConfiguration(config);
	}
	
	public void setClient(Client client){
		session.setClient(client);
	}
	
	/**
	 * set the hostname of this server
	 * 
	 * @param host as numeric IP address (TODO hostname lookup)
	 * @throws UnknownHostException
	 */
	public void setMyHost(String host)throws UnknownHostException{
		session.setMyHost(host);
	}
	
	public void setExportMode(boolean isExport)throws UnknownHostException{
		session.setSendMode(isExport);
	}
	
	public void run() throws Exception{
		session.setPeerHost(getRemoteHost());
		if(session.isSender())runExport();
		else runImport();
	}
	
	//send a file to the server
	protected void runExport()throws Exception{
		exchangeConnectionData();
		String exe=session.getConfiguration().getProperty(UDTSession.UDT_SEND);
		session.setExecutable(exe);
		session.doSendHandshake();
		//second part of handshake
		exchangeConnectionData();
	}
	
	//get a file from remote
	protected void runImport()throws Exception{
		String exe=session.getConfiguration().getProperty(UDTSession.UDT_RECEIVE);
		session.setExecutable(exe);
		session.startReceive();
		exchangeConnectionData();
		session.finishReceive();
	}
	
	private void exchangeConnectionData() throws IllegalStateException, BaseFault, UnknownHostException{
		//prepare params
		ExchangeParametersRequestDocument exch = ExchangeParametersRequestDocument.Factory.newInstance();
		exch.addNewExchangeParametersRequest().setClientHost(session.getMyHost());
		if(session.getMyPort()>-1){
			exch.getExchangeParametersRequest().setClientPort(BigInteger.valueOf(session.getMyPort()));
		}
		
		//send params and get feedback
		ExchangeParametersResponseDocument res = fts.initUDT(exch);
		
		//read information from the response document
		String peerHost = res.getExchangeParametersResponse().getServerHost();
		int peerPort = res.getExchangeParametersResponse().getServerPort().intValue();
		session.setPeerHost(peerHost);
		session.setPeerPort(peerPort);
	} 

	public void setLocalFile(String localFile){
		session.setLocalFile(localFile);
	}
	
	@Override
	public synchronized long getSourceFileSize() {
		if(datasize==-1){
			if(session.isSender()){
				try{
				   datasize=session.getLocalFileSize();
				}catch(ExecutionException e){
					LogUtil.logException("Could not get size of local file.",e,logger);
				}
			}
			else{
				datasize=super.getSourceFileSize();
			}
		}
		return datasize;
	}
	
	
	public String getRemoteHost()throws Exception{
		HostDocument o=HostDocument.Factory.parse(getResourceProperty(UDTFileTransfer.UDT_HOSTNAME));
		return o.getHost();
	}
	
	public boolean getRemoteHostIsSender()throws Exception{
		IsSenderDocument o=IsSenderDocument.Factory.parse(getResourceProperty(UDTFileTransfer.UDT_IS_SENDER));
		return o.getIsSender();
	}
	
	
	@Override
	public void readAllData(OutputStream os) throws Exception {
		//receive temp file and then copy it to the outstream...
		File tmpFile=File.createTempFile("unicore-udt", "tmp");
		tmpFile.deleteOnExit();
		setLocalFile(tmpFile.getAbsolutePath());
		run();
		//now copy to output
		FileInputStream fis=null;
		try{
			fis=new FileInputStream(tmpFile);
			copy(fis,os);
			os.flush();
		}finally{
			if(fis!=null)fis.close();
		}
	}
	
	@Override
	public void writeAllData(InputStream source) throws Exception {
		//spool data to local temp file
		File tmpFile=File.createTempFile("unicore-udt", "tmp");
		tmpFile.deleteOnExit();
		FileOutputStream fos=null;
		try{
			fos=new FileOutputStream(tmpFile);
			copy(source,fos);
			fos.flush();
		}finally{
			if(fos!=null)fos.close();
		}
		//now send the temp file
		setLocalFile(tmpFile.getAbsolutePath());
		run();
	}

	/**
	 * checks if an error occured and throws exception if it did
	 * @throws IOException
	 */
	public void checkError()throws IOException{
		try{
			session.checkError();
		}catch(ExecutionException ee){
			throw new IOException("TSI problem when checking for errors: "+ee.getMessage());
		}
	}
	
	@Override
	public synchronized void destroy()throws BaseFault,ResourceUnavailableFault,ResourceUnknownFault,ResourceNotDestroyedFault,ClientException{
		session.destroy();
		super.destroy();
	}

	@Override
	public long getTransferredBytes()throws Exception{
		if(session.isSender()){
			return super.getTransferredBytes();
		}
		try{
			long transferred=session.getLocalFileSize();
			TransferredBytesDocument tb=TransferredBytesDocument.Factory.newInstance();
			tb.setTransferredBytes(transferred);
			UpdateResourcePropertiesDocument srp=UpdateResourcePropertiesDocument.Factory.newInstance();
			srp.addNewUpdateResourceProperties().addNewUpdate().set(tb);
			getRP().UpdateResourceProperties(srp);
			return transferred;
		}catch(Exception ex){
			LogUtil.logException("Could not get size of local file.",ex,logger);
		}
		return -1;
	}
	
	@Override
	public ProtocolType.Enum getProtocol(){
		return ProtocolType.UDT;
	}
	
	private void copy(InputStream is, OutputStream os)throws IOException{
		byte[] buf=new byte[8192];
		while(true){
			int len=is.read(buf);
			if(len==-1)return;
			os.write(buf,0,len);
		}
	}
	
}
