package de.fzj.unicore.uas.fts.udt;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.UUID;

import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.ems.Action;
import de.fzj.unicore.xnjs.ems.ExecutionContext;
import de.fzj.unicore.xnjs.ems.ExecutionException;
import de.fzj.unicore.xnjs.ems.IExecutionContextManager;
import de.fzj.unicore.xnjs.tsi.TSI;
import de.fzj.unicore.xnjs.tsi.TSIBusyException;
import eu.unicore.security.Client;

/**
 * Utility to communicate with a UDT send/receive process on the TSI <br/>
 * 
 * Messages can be sent to the process using a file. Output from the process
 * can be read from standard output 
 * 
 * @author schuller 
 */
public class UDTCmdlineProcess {
	
	private final ExecutionContext ctx;
	private final Configuration config;
	private final Client client;
	private final String uuid;
	private boolean keepUspace;
	private final String commOutFileName;
	private final String commInFileName;
	
	/**
	 * creates a new process in an existing working
	 * 
	 * @param executable - the command to execute (including any commandline options)
	 * @param client
	 * @param config
	 * @param workingDirectory
	 * @throws Exception
	 */
	public UDTCmdlineProcess(String executable, Client client, Configuration config, String workingDirectory)throws Exception{
		IExecutionContextManager ecm=config.getExecutionContextMgr();
		this.config=config;
		this.client=client;
		this.uuid=UUID.randomUUID().toString();
		Action a=new Action();
		a.setClient(client);
		a.setUUID(uuid);
		commOutFileName="out."+uuid;
		commInFileName="in."+uuid;
		ctx=ecm.getContext(a);
		ctx.setExecutable(executable+" "+commOutFileName+" "+commInFileName);
		ctx.setStderr("stderr."+uuid);
		ctx.setStdout("stdout."+uuid);
		ctx.setInteractive(true);
		if(workingDirectory!=null)ctx.setWorkingDirectory(workingDirectory);
	}
	
	/**
	 * creates a new process in a new working directory
	 * @param executable
	 * @param client
	 * @param config
	 * @throws Exception
	 */
	public UDTCmdlineProcess(String executable, Client client, Configuration config)throws Exception{
		this(executable, client, config, null);
	}
	
	public ExecutionContext getExecutionContext(){
		return ctx;
	}
	
	/**
	 * async start of the process
	 * @throws Exception
	 */
	public void start() throws Exception{
		while(true){
			try{
				TSI tsi=config.getTargetSystemInterface(client);
				tsi.exec(ctx.getExecutable(),ctx);
				break;
			}catch(TSIBusyException tbe){
				try{Thread.sleep(1000);}catch(InterruptedException ie){}
			}
		}
	}
	
	/**
	 * get a line from the application's output file
	 * 
	 * @return
	 * @throws IOException
	 * @throws ExecutionException
	 */
	public String getOutputLine()throws IOException,ExecutionException,InterruptedException{
		TSI tsi=config.getTargetSystemInterface(client);
		String file=ctx.getWorkingDirectory()+commOutFileName;
		InputStream in=null;
		BufferedReader br=null;
		String outTag="OUT: ";
		try{
			//check if file exists
			while(tsi.getProperties(file)==null){
				Thread.sleep(200);
			}
			in=tsi.getInputStream(file);
			br=new BufferedReader(new InputStreamReader(in));
			while(true){
				String line=br.readLine();
				if(line!=null && !line.startsWith(outTag)){
					continue;
				}
				return line.substring(outTag.length());
			}
		}finally{
			if(in!=null)try{in.close();}catch(IOException ex){}
			if(br!=null)try{br.close();}catch(IOException ex){}
		}
		
	}
	
	/**
	 * checks the log file for any errors. If errors occured, an IOException is thrown.
	 *  
	 * @throws IOException - thrown if errors occured
	 * @throws ExecutionException - thrown if log could not be checked
	 */
	public String getLog()throws IOException,ExecutionException{
		StringBuilder result=new StringBuilder();
		TSI tsi=config.getTargetSystemInterface(client);
		String file=ctx.getWorkingDirectory()+commOutFileName;
		InputStream in=null;
		BufferedReader br=null;
		String lineSep=System.getProperty("line.separator", "/n");
		try{
			in=tsi.getInputStream(file);
			br=new BufferedReader(new InputStreamReader(in));
			String line=null;
			while(true){
				line=br.readLine();
				if(line==null)break;
				result.append(line).append(lineSep);	
			}
			return result.toString();
		}
		finally{
			if(in!=null)in.close();
			if(br!=null)br.close();
		}
	}
	
	/**
	 * send the given input to the process
	 * 
	 * @param input
	 * @throws IOException
	 * @throws ExecutionException
	 */
	public void sendInput(String input)throws IOException,ExecutionException{
		TSI tsi=config.getTargetSystemInterface(client);
		String file=ctx.getWorkingDirectory()+commInFileName;
		OutputStream os=null;
		try{
			os=tsi.getOutputStream(file);
			OutputStreamWriter osw=new OutputStreamWriter(os);
			osw.write(input);
			osw.flush();
		}finally{
			if(os!=null)os.close();
		}
		return;
	}
	
	/**
	 * Deletes the uspace, see {@link #setKeepUspace(boolean)}
	 * 
	 * @throws IOException
	 * @throws ExecutionException
	 * @throws TSIBusyException
	 */
	public void destroy()throws IOException, ExecutionException, TSIBusyException{
		if(keepUspace)return;
		//cleanup uspace
		TSI tsi=config.getTargetSystemInterface(client);
		tsi.rmdir(ctx.getWorkingDirectory());
	}

	public boolean isKeepUspace() {
		return keepUspace;
	}

	/**
	 * set the keepUspace flag. If "true", the uspace will not be deleted on destroy()
	 * @param keepUspace
	 */
	public void setKeepUspace(boolean keepUspace) {
		this.keepUspace = keepUspace;
	}
	
	
	/**
	 * checks the log file for any errors. If errors occured, an IOException is thrown.
	 *  
	 * @throws IOException - thrown if errors occured
	 * @throws ExecutionException - thrown if log could not be checked
	 */
	public void checkError()throws IOException,ExecutionException{
		TSI tsi=config.getTargetSystemInterface(client);
		String file=ctx.getWorkingDirectory()+commOutFileName;
		InputStream in=null;
		BufferedReader br=null;
		try{
			in=tsi.getInputStream(file);
			br=new BufferedReader(new InputStreamReader(in));
			String line=null;
			do{
				line=br.readLine();
				if(line!=null && line.startsWith("ERROR")){
					throw new IOException(line);
				}
			}while(line!=null);
			in.close();
			br.close();
		}
		finally{
			if(in!=null)in.close();
			if(br!=null)br.close();
		}
	}
	
}
