package de.fzj.unicore.uas.fts.udt;

import java.io.IOException;
import java.io.Serializable;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;
import org.unigrids.services.atomic.types.ProtocolType;

import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.UDTClient;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.ems.ExecutionException;
import de.fzj.unicore.xnjs.io.XnjsFile;
import de.fzj.unicore.xnjs.tsi.TSI;
import eu.unicore.security.Client;

/**
 * stores the state of a UDT file transfer <br/> 
 * 
 * In UDT we have two parties. Party "Sender" where the source file is located, and Party "Receiver"
 * where the file is transferred to.<br/>
 * 
 * We have two corresponding command-line apps "send" and "receive", which are used to do the 
 * UDP holepunching handshake, and then transfer the file. The proceeding is as follows:
 * 
 * <ul>
 *  <li>the "receive" app is invoked: <code>recv [sender_ip] [target_filename]</code></li>
 *  <li>the "receive" app outputs the port [receiver_port] it uses
 *  <li>the "send" app is invoked: <code>send [receiver_ip] [receiver_port][source_filename]</code></li>
 *  <li>the "send" app outputs the port [sender_port] it uses</li>
 *  <li>the [sender_port] is written to the input of the "receive" process</li>
 *  <li>the data is now transferred</li>
 * </ul>
 * 
 * @author oistrez
 * @author schuller
 */
public class UDTSession implements Serializable{

	private static final Logger log=LogUtil.getLogger(LogUtil.SERVICES,UDTSession.class);
	
	private static final long serialVersionUID = 1L;

	private String senderHost;
	private int senderPort=-1;
	
	private String receiverHost;
	private int receiverPort=-1;

	private boolean appendToFile=false;
	
	public boolean isAppendToFile() {
		return appendToFile;
	}

	public void setAppendToFile(boolean appendToFile) {
		this.appendToFile = appendToFile;
	}

	/**
	 * the state of this session
	 */
	public static enum State {
		 NEW,
		 EXCHANGE_PARAMETERS,
		 RUNNING,
		 DONE,
		 FAILED
	}
	/**
	 * UAS property for defining the location of the UDT "send" application
	 */
	public static final String UDT_SEND="uas.udt.send";
	
	/**
	 * UAS property for defining the location of the UDT "receive" application
	 */
	public static final String UDT_RECEIVE="uas.udt.receive";


	/**
	 * property for defining the server host where UDT is running 
	 * (i.e. usually the TSI login node)<br/>
	 * (default: "127.0.0.1") <br/>
	 * TODO this should be inferred dynamically
	 */
	public static final String UDT_HOST="XNJS.udt.host";
	
	//"true" means the sender is on this machine (where the UDTSession object is) 
	private boolean isSender = true;
	
	/**
	 * the path to the file that is on this machine
	 */
	private String localFilePath = "";
	
	private volatile boolean ready=true;
	
	private State state=State.NEW;
	
	private UDTCmdlineProcess process;
	
	private Client client;
	
	private Configuration config;
	
	private String executable;
	
	static{	
		initUDT();
	}

	private static synchronized void initUDT(){
		//register UDT client with UNICORE
		StorageClient.registerClient(ProtocolType.UDT, UDTClient.class);
	}

	public boolean isReady(){
		return ready;
	} 
	
	//the reserved port for the UDT connection
	public int getMyPort() {
		if(isSender){
			return senderPort;
		}
		else{
			return receiverPort;
		}
	}
	
	//the local IP address for the UDT connection
	public String getMyHost() throws NullPointerException, UnknownHostException{
		if(isSender){
			return senderHost;
		}
		else{
			return receiverHost;
		}
	}

	public void setMyPort(int port){
		if(isSender){
			senderPort=port;
		}
		else{
			receiverPort=port;
		}
	}
	
	public void setMyHost(String host)throws UnknownHostException{
		if(isSender){
			senderHost=host;
		}
		else{
			receiverHost=host;
		}
	}

	
	public void setPeerPort(int port){
		if(isSender){
			receiverPort=port;
		}
		else{
			senderPort=port;
		}
	}
	
	//the remote port of the UDP connection
	public int getPeerPort() {
		if(isSender){
			return receiverPort;
		}
		else{
			return senderPort;
		}
	}
	public void setPeerHost(String host)throws UnknownHostException{
		if(isSender){
			receiverHost=host;
		}
		else{
			senderHost=host;
		}
	}
	
	//the local IP address for the UDT connection
	public String getPeerHost() throws NullPointerException, UnknownHostException{
		if(isSender){
			return receiverHost;
		}
		else{
			return senderHost;
		}
	}
	
	//the local file path
	public void setLocalFile(String path){
		this.localFilePath=path;
	}

	/**
	 * sets the mode. If <code>true</code>, this machine <em>sends</em> the file,
	 * if <code>false</code>, this machine <em>receives</em> the file
	 * 
	 * @param sendMode
	 */
	public void setSendMode(boolean sendMode) {
		this.isSender = sendMode;
	}

	public long getLocalFileSize()throws ExecutionException{
		TSI tsi=config.getTargetSystemInterface(client);
		String path="/.";
		String target=localFilePath;
		//if target contains separator char, split off the filename part
		int i=localFilePath.lastIndexOf('/');
		if(i!=-1){
			target=localFilePath.substring(i+1);
			path=localFilePath.substring(0, i);
		}
		XnjsFile[] files=tsi.ls(path);
		XnjsFile file=null;
		for(XnjsFile f: files){
			if(f.getPath().endsWith(target)){
				file=f;
				break;
			}
		}
		return file!=null?file.getSize():-1;
	}
	
	public boolean isSender() {
		return isSender;
	}

	public String getLocalFilePath() {
		return localFilePath;
	}

	public void setLocalFilePath(String localFilePath) {
		this.localFilePath = localFilePath;
	}

	
	public void startReceive()throws IllegalStateException, Exception{
		if(log.isDebugEnabled()){
			log.debug("startReceive: "+toString());
		}
		//some sanity checks first
		if(!State.NEW.equals(state)){
			throw new IllegalStateException("UDT Session in inconsistent state");
		}
		ready=false;
		if(isSender){
			//nothing to do here...
			throw new IllegalStateException("UDT session in inconsistent state");
		}
		String mode=appendToFile?"A":"O";
		//setup interactive process for "receive"
		process=new UDTCmdlineProcess(executable+" "+getPeerHost()+" "+localFilePath+" "+mode,client,config);
		process.setKeepUspace(true);
		process.start();
		log.info("Started UDT receive");
		//get the port from stdout
		String port=process.getOutputLine().trim();
		setMyPort(Integer.parseInt(port));
		state=State.EXCHANGE_PARAMETERS;
		ready=true;
		if(log.isDebugEnabled()){
			log.debug("after startReceive: "+toString());
		}
	}

	public void finishReceive()throws IllegalStateException, Exception{
		if(log.isDebugEnabled()){
			log.debug("finishReceive: "+toString());
		}
		//some sanity checks first
		if(!State.EXCHANGE_PARAMETERS.equals(state)){
			throw new IllegalStateException("UDT Session in inconsistent state");
		}
		ready=false;
		if(isSender){
			//nothing to do here...
			throw new IllegalStateException("UDT session in inconsistent state");
		}
		//send the peer port to stdin
		process.sendInput(Integer.toString(getPeerPort())+"\n");
		state=State.RUNNING;
		ready=true;
		if(log.isDebugEnabled()){
			log.debug("after finishReceive: "+toString());
		}
	}

	public void doSendHandshake()throws IllegalStateException, Exception{
		if(log.isDebugEnabled()){
			log.debug("sendHandshake: "+toString());
		}
		if(!State.NEW.equals(state)){
			throw new IllegalStateException("UDT Session in inconsistent state");
		}
		ready=false;
		if(isSender){
			//setup interactive process for "send"
			process=new UDTCmdlineProcess(executable+" "+getPeerHost()+" "+getPeerPort()
					+" "+localFilePath,client,config);
			process.setKeepUspace(true);
			process.start();
			log.info("Started UDT send process");
			//get the port from stdout
			String port=process.getOutputLine().trim();
			setMyPort(Integer.parseInt(port));
			log.info("UDT sender ready at port <"+port+">");
		}
		state=State.EXCHANGE_PARAMETERS;
		ready=true;
		if(log.isDebugEnabled()){
			log.debug("after sendHandshake: "+toString());
			log.debug("Process log after sendHandshake: "+process.getLog());
		}
	}

	public void setClient(Client client) {
		this.client = client;
	}

	public void setConfiguration(Configuration config) {
		this.config = config;
	}

	public Configuration getConfiguration(){
		return config;
	}

	public String getExecutable() {
		return executable;
	}

	public void setExecutable(String executable) {
		this.executable = executable;
	}

	public State getState() {
		return state;
	}
	
	public void destroy(){
		if(log.isDebugEnabled()){
			log.debug("destroySession: "+toString());
		}
		try{
			if(process!=null)process.destroy();
		}catch(Exception e){
			LogUtil.logException("Error destroying UDT session.",e,log);
		}
	}
	
	public void checkError()throws IOException,ExecutionException{
		process.checkError();
	}
	
	public String toString(){
		StringBuilder sb=new StringBuilder();
		sb.append(super.toString()+" state: "+state);
		try{
			sb.append("\n");
			sb.append("My address: ");
			sb.append(getMyHost());
			sb.append(":");
			sb.append(getMyPort());
		}catch(UnknownHostException uhe){
			LogUtil.logException("My host is unknown",uhe,log);
		}
		try{
			sb.append("\n");
			sb.append("Peer address: ");
			sb.append(getPeerHost());
			sb.append(":");
			sb.append(getPeerPort());
		}catch(UnknownHostException uhe){
			LogUtil.logException("Peer host is unknown",uhe,log);
		}
		sb.append("\n");
		sb.append("Local file: "+getLocalFilePath());
		return sb.toString();
	}
	
	/**
	 * get the numeric IP address of the local host
	 * @return "127.0.0.1" in IPv4 or "::1" in IPv6
	 */
	public static String getLocalhostAddress(){
		try{
			if(InetAddress.getLocalHost() instanceof Inet6Address){
				return "::1";
			}
			else return "127.0.0.1";
		}catch(Exception ex){
			return "127.0.0.1";
		}
	}
}